import pkgutil
from itertools import repeat
from sympy import flatten


def string_list(strlist):
    return ', '.join(strlist) if isinstance(strlist, list) else strlist


class MetricData (dict):
    def __init__(self, doc):
        import yaml
        super().__init__(yaml.safe_load(doc))
        doc_items = [string_list(val) for val in self.values()]
        self.__doc__ = '\n'.join(doc_items)


metric_data = []

for module_info in pkgutil.iter_modules(__path__, __name__ + '.'):
    module = __import__(module_info.name, fromlist='dummy')
    if module.__doc__:
        entry = MetricData(module.__doc__.lower())
        entry['metric'] = module.metric
        entry['coords'] = module.coords
        entry['variables'] = module.variables
        entry['functions'] = module.functions
        metric_data.append(entry)


def compare_filter(entries, value, key):
    if not value:
        return entries
    func = lambda entry: value.lower() in string_list(entry.get(key, ''))
    return filter(func, entries)


def find(sub=None, symmetries=None, coords=None, notes=None):
    entries = metric_data

    if symmetries:
        for symmetry in flatten([symmetries]):
            entries = compare_filter(entries, symmetry, 'symmetry')

    if coords:
        entries = compare_filter(entries, coords, 'coordinates')

    if notes:
        for note in flatten([notes]):
            entries = compare_filter(entries, note, 'notes')

    if sub is None:
        retval = map(MetricData.__getitem__, entries, repeat('name'))
    else:
        retval = [entry['name'] for entry in entries if sub.lower() in entry.__doc__]

    retval = list(set(retval))
    retval.sort()
    return retval


def data(name, coords=None, notes=None):
    entries = [metric for metric in metric_data if metric['name'] == name.lower()]
    if not entries:
        raise KeyError('metric {} not found'.format(name))

    if coords:
        entries = compare_filter(entries, coords, 'coordinates')

    if notes:
        for note in flatten([notes]):
            entries = compare_filter(entries, note, 'notes')

    entries = list(entries)
    if len(entries) == 1:
        return entries[0]
    return entries


def coordinate_types(name, notes=None):
    entries = data(name, notes=notes)
    if not isinstance(entries, list):
        return entries.get('coordinates')
    return set([entry.get('coordinates') for entry in entries])


def variations(name, coords=None):
    entries = data(name, coords=coords)
    if not isinstance(entries, list):
        return entries.get('notes')
    return set([entry.get('notes') for entry in entries])



"""
Name: Novotny and Horsky Vacuum
References: Novotny et al., Can. J. Phys., v24, p718, (1974)
Coordinates: Cartesian
Symmetry: Planar
"""
from sympy import Rational, cos, diag, sin, symbols

coords = symbols('t x y z', real=True)
variables = symbols('a', constant=True)
functions = ()
t, x, y, z = coords
a = variables
metric = diag(
    -cos(z) / sin(z) ** Rational(2, 3),
    sin(z) ** Rational(4, 3),
    sin(z) ** Rational(4, 3),
    1 / a ** 2,
)



"""
Name: Vaidya
References: Stephani (13.20) p158
Coordinates: Eddington-Finkelstein
Notes:
    - Cosmological Constant
    - Ingoing Coordinates
"""
from sympy import Function, diag, sin, symbols

coords = symbols('r v theta phi', real=True)
variables = symbols('Lambda', constant=True)
functions = symbols('M', cls=Function)
r, v, th, ph = coords
Lambda = variables
M = functions
metric = diag(
    0, -(1 - 2 * M(v) / r - Lambda * r ** 2 / 3), r ** 2, r ** 2 * sin(th) ** 2
)
metric[0, 1] = metric[1, 0] = 1



"""
Name: Schwarzschild
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
Notes: Tolman-Bondi Dust Limiting Case
"""
from sympy import Rational, diag, sin, symbols

coords = symbols('t r theta phi', real=True)
variables = symbols('M', constant=True)
functions = ()
t, r, th, ph = coords
M = variables
expr = M * (r - t) ** 2
metric = diag(
    -1,
    Rational(4, 9) * M / expr ** Rational(1, 3),
    expr ** Rational(2, 3),
    expr ** Rational(2, 3) * sin(th) ** 2,
)



"""
Name: Einstein-Maxwell Field
References: Stephani (20.9a) p221
Coordinates: Cylindrical
Symmetry:
    - Cylindrical
    - Static
Notes: Angular Magnetic Field
"""
from sympy import cosh, diag, log, symbols

coords = symbols('t rho phi z', real=True)
variables = symbols('a b m', constant=True)
functions = ()
t, rh, ph, z = coords
a, b, m = variables
expr1 = cosh(log(a * rh ** m)) ** 2
expr2 = rh ** (2 * m ** 2) * b ** 2 * expr1
metric = diag(-expr2, expr2, rh ** 2 * b ** 2 * expr1, 1 / (b ** 2 * expr1))



"""
Name: Koutras-McIntosh
References: Koutras et al., Class. Quantum Grav., v13, pL47, (1996)
"""
from sympy import Function, symbols, zeros

coords = symbols('w u x y', real=True)
variables = symbols('a b', constant=True)
functions = symbols('f', cls=Function)
w, u, x, y = coords
a, b = variables
f = functions
metric = zeros(4)
metric[1, 1] = -(2 * f(u) * (a * x + b) * (x ** 2 + y ** 2) - a ** 2 * w ** 2)
metric[2, 2] = 1
metric[3, 3] = 1
metric[0, 1] = metric[1, 0] = -(a * x + b)
metric[1, 2] = metric[2, 1] = a * w



"""
Name: Levi-Civita Vacuum
References: Stephani (20.8) p221
Coordinates: Cylindrical
"""
from sympy import diag, symbols

coords = symbols('t rho phi z', real=True)
variables = symbols('m', constant=True)
functions = ()
t, rh, ph, z = coords
m = variables
metric = diag(
    -rh ** (2 * m),
    rh ** (2 * (m ** 2 - m)),
    rh ** (2 * (1 - m)),
    rh ** (2 * (m ** 2 - m)),
)



"""
Name: Schwarzschild
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
Notes: Interior
"""
from sympy import diag, sin, sqrt, symbols

coords = symbols('t r theta phi', real=True)
variables = symbols('A B R', constant=True)
functions = ()
t, r, th, ph = coords
A, B, R = variables
metric = diag(
    -(A - B * sqrt(1 - r ** 2 / R ** 2)) ** 2,
    1 / (1 - r ** 2 / R ** 2),
    r ** 2,
    r ** 2 * sin(th) ** 2,
)



"""
Name: de Sitter
References: Hawking and Ellis p125
Symmetry: Maximal
"""
from sympy import cosh, diag, sin, symbols

coords = symbols('t chi theta phi', real=True)
variables = symbols('alpha', constant=True)
functions = ()
t, ch, th, ph = coords
al = variables
expr = al ** 2 * cosh(t / al) ** 2
metric = diag(-1, expr, expr * sin(ch) ** 2, expr * sin(ch) ** 2 * sin(th) ** 2)



"""
Name: Heintzmann Perfect Fluid
References: Heintzmann, Z. Phys., v228, p489-493, (1969)
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
"""
from sympy import diag, sin, symbols

coords = symbols('t r theta phi', real=True)
variables = symbols('A a K', constant=True)
functions = ()
t, r, th, ph = coords
A, a, K = variables
metric = diag(
    -A ** 2 * (1 + a * r ** 2) ** 3, (1 + a * r ** 2) / K, r ** 2, r ** 2 * sin(th) ** 2
)



"""
Name: Davidson Perfect Fluid
References: Davidson, J. Math. Phys., v32, p1560, (1991)
Coordinates: Cylindrical
Symmetry: Cylindrical
"""
from sympy import Rational, diag, symbols

coords = symbols('t r phi z', real=True)
variables = ()
functions = ()
t, r, phi, z = coords
expr = (1 + r ** 2) ** Rational(2, 5)
metric = diag(
    -expr ** 3,
    t ** Rational(4, 3) * expr,
    t ** Rational(4, 3) * r ** 2 / expr,
    t ** Rational(-2, 3) / expr,
)



"""
Name: McLenaghan-Tariq-Tupper
References:
    - McLenaghan, J. Math. Phys., v16, p11, (1975)
    - Tupper, Gen. Rel. Grav., v7, p479, (1976)
    - Stephani (10.21) p121
"""
from sympy import diag, symbols

coords = symbols('t x y phi', real=True)
variables = symbols('a', constant=True)
functions = ()
t, x, y, ph = coords
a = variables
metric = diag(-1, a ** 2 / x ** 2, a ** 2 / x ** 2, x ** 2 - 4 * y ** 2)
metric[0, 3] = metric[3, 0] = 2 * y



from sympy import diag, sin, symbols

from riccipy.metric import *
from riccipy.partial import *
from riccipy.tensor import *


def _generate_schwarzschild():
    coords = symbols('t r theta phi', real=True)
    t, r, th, ph = coords
    schw = diag(1 - 1 / r, -1 / (1 - 1 / r), -r ** 2, -r ** 2 * sin(th) ** 2)
    g = SpacetimeMetric('g', coords, schw, timelike=True)
    mu, nu = indices('mu nu', g)
    return (coords, t, r, th, ph, schw, g, mu, nu)


def test_DiffOperator():
    x, y = symbols('x y')
    dx = DiffOperator(x)
    assert (0 * dx) == 0
    expr = dx * (1 - 1 / x)
    assert expr == 1 / x ** 2
    expr = (1 - 1 / x) * dx
    assert isinstance(expr, DiffOperator)
    assert (expr * (x ** 3 / 3)).equals(x * (x - 1))
    dy = DiffOperator(y)
    res = dx * dy
    assert res.args == (x, y)


def test_PartialDerivative():
    (coords, t, r, th, ph, schw, g, mu, nu) = _generate_schwarzschild()
    d = g.partial
    assert isinstance(d, PartialDerivative)
    assert isinstance(d, Tensor)
    assert d.commutes_with(d) == 0
    assert d.covar == (-1,)
    assert all([isinstance(dx, DiffOperator) for dx in d.as_array()])



"""
Name: Minkowski
Coordinates: 'Null'
Symmetry: Maximal
"""
from sympy import Rational, sin, symbols, zeros

coords = symbols('u v theta phi', real=True)
variables = symbols('a b c', constant=True)
functions = ()
u, v, th, ph = coords
a, b, c = variables
metric = zeros(4)
metric[2, 2] = (a * u - Rational(1, 2) * v / a + c) ** 2
metric[3, 3] = (a * u - Rational(1, 2) * v / a + c) ** 2 * sin(th) ** 2
metric[0, 1] = metric[1, 0] = -1



"""
Name: Harrison
References:
    - Harrison, Phys. Rev., v116, p1285, (1959)
    - d'Inverno et al., J. Math. Phys., v12, p1258, (1971)
Notes:
    - Kinnersley Class IV.B
    - C = 0
"""
from sympy import Rational, diag, sin, sqrt, symbols

coords = symbols('x_0:4', real=True)
variables = ()
functions = ()
x0, x1, x2, x3 = coords
metric = diag(
    -x1 * sin(x3),
    1 / (sqrt(x1 * sin(x3))),
    x1 * sin(x3),
    x1 ** Rational(3, 2) / sqrt(sin(x3)),
)



"""
Name: Ellis and MacCallum Dust
References:
    - Ellis et al., Commun. Math. Phys., v12, p108, (1969)
    - Dunn et al., Astrophys. J., v204, p322, (1976)
    - Evans, Mon. Not. R. Ast. Soc., v183, p727, (1978)
    - Stephani (12.25) p150
Coordinates: Cartesian
"""
from sympy import diag, exp, symbols

coords = symbols('t x y z', real=True)
variables = ()
functions = ()
t, x, y, z = coords
metric = diag(-1, 16 * t ** 2, t * exp(-4 * x), t * exp(4 * x))



"""
Name: Durgapal
References: Durgapal, J. Phys. A, v15, p2637-2644, (1982)
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
Notes: n = 5
"""
from sympy import Rational, sin, symbols, zeros

coords = symbols('t r theta phi', real=True)
variables = symbols('A C K', constant=True)
functions = ()
t, r, th, ph = coords
A, C, K = variables
metric = zeros(4)
metric[0, 0] = -A ** 2 * (1 + C * r ** 2) ** 5
metric[1, 1] = (1 + C * r ** 2) ** 3 / (
    1
    - C * r ** 2 * (309 + 54 * C * r ** 2 + 8 * C ** 2 * r ** 4) / 112
    + K * C * r ** 2 / (1 + 6 * C * r ** 2) ** Rational(1, 3)
)
metric[2, 2] = r ** 2
metric[3, 3] = r ** 2 * sin(th) ** 2



"""
Name: Tolman
References: Tolman, Phys. Rev., v55, p363-373, (1939)
Coordinates: Spherical
Notes: Type IV
"""
from sympy import diag, sin, symbols

coords = symbols('t r theta phi', real=True)
variables = symbols('A B R', constant=True)
functions = ()
t, r, th, ph = coords
A, B, R = variables
metric = diag(
    -B ** 2 * (1 + r ** 2 / A ** 2),
    (1 + 2 * r ** 2 / A ** 2) / ((1 + r ** 2 / A ** 2) * (1 - r ** 2 / R ** 2)),
    r ** 2,
    r ** 2 * sin(th) ** 2,
)



"""
Name: Domain Wall
Coordinates: Cartesian
"""
from sympy import diag, exp, symbols

coords = symbols('t x y z', real=True)
variables = (symbols('k', constant=True),)
functions = ()
t, x, y, z = coords
k, = variables
expr1 = (1 - k * z) ** 2
expr2 = exp(2 * k * t)
metric = diag(-expr1, expr1 * expr2, expr1 * expr2, 1)



"""
Name: Cross product of Constant Curvature Subspaces
References: Stephani (10.8) p118
Notes:
    - Temporal Sine
    - Spatial Sine
"""
from sympy import diag, sin, symbols

coords = symbols('t x y z', real=True)
variables = symbols('A B', constant=True)
functions = ()
t, x, y, z = coords
A, B = variables
metric = diag(-B ** 2 * sin(z) ** 2, A ** 2, A ** 2 * sin(x) ** 2, B ** 2)



"""
Name: Bianchi II
Coordinates: Cartesian
"""
from sympy import Function, diag, exp, symbols

coords = symbols('t x y z', real=True)
variables = ()
functions = symbols('alpha', cls=Function)
t, x, y, z = coords
al = functions
metric = diag(1, exp(-2 * al(t)), exp(al(t)), exp(al(t)))



"""
Name: Datta
References:
    - Datta, Nuovo Cim., v36, p109
    - Stephani (11.60) p137
Coordinates: Cartesian
Notes: Type 2
"""
from sympy import diag, symbols

coords = symbols('t x y z', real=True)
variables = symbols('C_1 C_2 mu', constant=True)
functions = ()
t, x, y, z = coords
C1, C2, mu = variables
expr = (C1 * t ** mu + C2 / t ** mu) ** 2
metric = diag(
    -expr * t ** (2 * mu ** 2), 1 / expr, expr * t ** (2 * mu ** 2), t ** 2 * expr
)



"""
Name: Durgapal
References: Durgapal, J. Phys. A, v15, p2637-2644, (1982)
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
Notes: n = 3
"""
from sympy import sin, sqrt, symbols, zeros

coords = symbols('t r theta phi', real=True)
variables = symbols('A C K', constant=True)
functions = ()
t, r, th, ph = coords
A, C, K = variables
metric = zeros(4)
metric[0, 0] = -A ** 2 * (1 + C * r ** 2) ** 3
metric[1, 1] = (
    2
    * (1 + C * r ** 2)
    * sqrt(1 + 4 * C * r ** 2)
    / ((2 - C * r ** 2) * sqrt(1 + 4 * C * r ** 2) + 2 * K * C * r ** 2)
)
metric[2, 2] = r ** 2
metric[3, 3] = r ** 2 * sin(th) ** 2



"""
Name: Bertotti-Robinson
References:
    - Bertotti, Phys. Rev., v116, p1331, (1959)
    - Lovelock, Commun. Math. Phys., v5, p257, (1967)
    - Dolan, Commun. Math. Phys., v9, p161, (1968)
    - Stephani (10.19) p120
Notes: Temporal Hyperbolic Sine
"""
from sympy import diag, sin, sinh, symbols

coords = symbols('t x theta phi', real=True)
variables = symbols('k', constant=True)
functions = ()
t, x, th, ph = coords
k = variables
metric = diag(-k ** 2 * sinh(x) ** 2, k ** 2, k ** 2, k ** 2 * sin(th) ** 2)



"""
Name: Harrison
References:
    - Harrison, Phys. Rev., v116, p1285, (1959)
    - d'Inverno et al., J. Math. Phys., v12, p1258, (1971)
"""
from sympy import Rational, cosh, symbols, zeros

coords = symbols('x_0:4', real=True)
variables = symbols('a', constant=True)
functions = ()
x0, x1, x2, x3 = coords
a = variables
expr = x1 ** 2 * cosh(x3) ** Rational(3, 2)
metric = zeros(4)
metric[0, 0] = x0 ** Rational(3, 2) * expr
metric[1, 1] = a ** 2 * x0 ** 3 * x1 * cosh(x3) ** 3
metric[2, 2] = 1 / (x0 * x1 * cosh(x3))
metric[3, 3] = x0 ** Rational(7, 2) * expr



"""
Name: Kasner Vacuum
Coordinates: Cartesian
Symmetry: Axial
"""
from sympy import diag, symbols

coords = symbols('t x y z', real=True)
variables = ()
functions = ()
t, x, y, z = coords
metric = diag(-1, t ** 2, 1, 1)



"""
Name: Bertotti-Robinson
References:
    - Bertotti, Phys. Rev., v116, p1331, (1959)
    - Bertotti, Commun. Math. Phys., v5, p257, (1967)
    - Robinson, Commun. Math. Phys., v9, p161, (1968)
    - Stephani (10.16) p120
Coordinates: Cartesian
"""
from sympy import diag, symbols

coords = symbols('t x y z', real=True)
variables = ()
functions = ()
t, x, y, z = coords
metric = diag(-(1 + z ** 2), 1 - y ** 2, 1 / (1 - y ** 2), 1 / (1 + z ** 2))



"""
Name: Schwarzschild-de Sitter
Coordinates: Spherical
Symmetry: Spherical
Notes: Cosmological Constant
"""
from sympy import diag, sin, symbols

coords = symbols('t r theta phi', real=True)
variables = symbols('M Lambda', constant=True)
functions = ()
t, r, th, ph = coords
M, La = variables
expr = 1 - 2 * M / r - La * r ** 2 / 3
metric = diag(-expr, 1 / expr, r ** 2, r ** 2 * sin(th) ** 2)



"""
Name: Datta
References:
    - Datta, Nuovo Cim., v36, p109
    - Stephani (11.60) p137
Coordinates: Cartesian
Notes: Type 1
"""
from sympy import diag, symbols

coords = symbols('t x y z', real=True)
variables = symbols('a b', constant=True)
functions = ()
t, x, y, z = coords
a, b = variables
metric = diag(-1 / (b / t - a / t ** 2), b / t - a / t ** 2, t ** 2, t ** 2)



"""
Name: Flat Boost Isotropy
References: Stephani (11.16) p128
Symmetry: Boost Rotation
"""
from sympy import Function, diag, symbols

coords = symbols('t w x y', real=True)
variables = ()
functions = symbols('alpha beta', cls=Function)
t, w, x, y = coords
al, be = functions
metric = diag(-be(w) ** 2 * y ** 2, 1, al(w) ** 2, be(w) ** 2)



"""
Name: Kowalczynski and Plebanski
References:
    - Kowalczynski et al., Int. J. Theor. Phys., v16, p371, (1977)
    - Stephani (27.57) p297
Coordinates: Cartesian
"""
from sympy import diag, symbols

coords = symbols('t x y z', real=True)
variables = symbols('a b c d', constant=True)
functions = ()
t, x, y, z = coords
a, b, c, d = variables
expr1 = a * z ** 2 + b
expr2 = -2 * d ** 2 * x ** 2 + c * x - a
metric = diag(
    -2 * expr1 / x ** 2, 2 / (expr2 * x ** 4), 2 * expr2, 2 / (expr1 * x ** 2)
)



"""
Name: McVittie Einstein-Maxwell Field
References:
    - McVittie, Prog. Roy. Soc. Lond., v124, p366, (1929)
    - Stephani (13.26) p158
Coordinates: Cartesian
"""
from sympy import diag, symbols

coords = symbols('t x y z', real=True)
variables = symbols('M Q', constant=True)
functions = ()
t, x, y, z = coords
M, Q = variables
metric = diag(-(M / z + Q ** 2 / z ** 2), z ** 2, z ** 2, 1 / (M / z + Q ** 2 / z ** 2))



"""
Name: Faulkes Perfect Fluid
References: Faulkes, Prog. Theor. Phys., v42, p1139-1142, (1969)
Coordinates: Spherical
Symmetry: Spherical
"""
from sympy import Function, diag, sin, symbols

coords = symbols('t r theta phi', real=True)
variables = symbols('k', constant=True)
functions = symbols('E', cls=Function)
t, r, th, ph = coords
k = variables
E = functions
expr = (E(t) - (r / (2 * k))) ** 2
metric = diag(
    -E(t) ** 2 / expr, expr ** 2, expr ** 2 * r ** 2, expr ** 2 * r ** 2 * sin(th) ** 2
)



"""
Name: Levi-Civita
References: Stephani (Table 16.2) p188
Coordinates: Cylindrical
Notes: Class A3
"""
from sympy import diag, symbols

coords = symbols('t r phi z', real=True)
variables = ()
functions = ()
t, r, ph, z = coords
metric = diag(-1 / z, z ** 2, z ** 2 * r ** 2, z)



"""
Name: Harrison
References:
    - Harrison, Phys. Rev., v116, p1285, (1959)
    - d'Inverno et al., J. Math. Phys., v12, p1258, (1971)
Notes:
    - Kinnersley Class IV.B
    - C = 1/2
    - Hyperbolic Cosine
"""
from sympy import cosh, diag, symbols

coords = symbols('x_0:4', real=True)
variables = ()
functions = ()
x0, x1, x2, x3 = coords
metric = diag(
    -1 / (1 - x3 ** 2) ** 2,
    cosh(2 * x0) ** 2 / (1 - x3 ** 2) ** 2,
    x3 ** 2,
    1 / (1 - x3 ** 2) ** 4,
)



"""
Name: Anti-de Sitter
References: Hawking and Ellis (5.9) p131
Coordinates: Spherical
Symmetry: Maximal
Notes: Static
"""
from sympy import cosh, diag, sin, sinh, symbols

coords = symbols('t r theta phi', real=True)
variables = ()
functions = ()
t, r, th, ph = coords
metric = diag(-cosh(r) ** 2, 1, sinh(r) ** 2, sinh(r) ** 2 * sin(th) ** 2)



"""
Name: Reissner-Nordstrom Electro-Vacuum
References:
    - Reissner, Ann. Phys., v50, p106, (1916)
    - Stephani (Table 13.1) p157
Coordinates: Spherical
Symmetry: Spherical
Notes: Cosmological Constant
"""
from sympy import diag, sin, symbols

coords = symbols('t r theta phi', real=True)
variables = symbols('M Q Lambda', constant=True)
functions = ()
t, r, th, ph = coords
M, Q, La = variables
expr = 1 - 2 * M / r + Q ** 2 / r ** 2 - La * r ** 2 / 3
metric = diag(-expr, 1 / expr, r ** 2, r ** 2 * sin(th) ** 2)



"""
Name: Bondi
References: Bondi, Proc. Roy. Soc. Lond. A, v282, p303, (1964)
Coordinates: Spherical
Symmetry: Spherical
Notes: Outgoing Coordinates
"""
from sympy import Function, diag, sin, symbols

coords = symbols('r v theta phi', real=True)
variables = ()
functions = symbols('C M', cls=Function)
r, v, th, ph = coords
C, M = functions
metric = diag(0, -C(r, v) ** 2 * (1 - 2 * M(r, v) / r), r ** 2, r ** 2 * sin(th) ** 2)
metric[0, 1] = metric[1, 0] = -C(r, v)



"""
Name: Schwarzschild
References:
    - Eddington, Nature, v113, p192, (1924)
    - Finkelstein, Phys. Rev., v110, p965, (1958)
    - Stephani (13.23) p158
Coordinates: Eddington-Finkelstein
Symmetry:
    - Spherical
    - Static
Notes: Ingoing Coordinates
"""
from sympy import diag, sin, symbols

coords = symbols('r v theta phi', real=True)
variables = symbols('M', constant=True)
functions = ()
r, v, th, ph = coords
M = variables
metric = diag(0, -(1 - 2 * M / r), r ** 2, r ** 2 * sin(th) ** 2)
metric[0, 1] = metric[1, 0] = 1



"""
Name: Taub Vacuum
References: Taub, Ann. Math., v53, p473, (1951)
Coordinates: Cartesian
Symmetry: Planar
"""
from sympy import diag, sqrt, symbols

coords = symbols('t x y z', real=True)
variables = ()
functions = ()
t, x, y, z = coords
metric = diag(-1 / sqrt(z), z, z, 1 / sqrt(z))



"""
Name: Friedman-Robertson-Walker Dust
References:
    - Landau-Lifshitz (112.4), (112.9), (112.10)
    - Stephani (12.3a) p122
    - Hawking and Ellis ch5.3
Notes: Closed
"""
from sympy import cos, diag, sin, symbols

coords = symbols('nu chi theta phi', real=True)
variables = symbols('A_0', constant=True)
functions = ()
nu, ch, th, ph = coords
A0 = variables
expr = A0 ** 2 * (1 - cos(nu)) ** 2
metric = diag(-expr, expr, expr * sin(ch) ** 2, expr * sin(ch) ** 2 * sin(th) ** 2)



"""
Name: Bertotti-Robinson
References:
    - Bertotti, Phys. Rev., v116, p1331, (1959)
    - Lovelock, Commun. Math. Phys., v5, p257, (1967)
    - Dolan, Commun. Math. Phys., v9, p161, (1968)
    - Stephani (32.95) p372
Notes: Cosine
"""
from sympy import cos, diag, symbols

coords = symbols('t x y z', real=True)
variables = symbols('omega', constant=True)
functions = ()
t, x, y, z = coords
om = variables
metric = diag(-1, cos(om * x) ** 2, cos(om * t) ** 2, 1)



"""
Name: Schwarzschild
Coordinates: Cartesian
Symmetry:
    - Spherical
    - Static
Notes: Isotropic Coordinates
"""
from sympy import Rational, diag, sqrt, symbols

coords = symbols('t x y z', real=True)
variables = symbols('M', constant=True)
functions = ()
t, x, y, z = coords
M = variables
expr1 = sqrt(x ** 2 + y ** 2 + z ** 2)
expr2 = (1 + Rational(1, 2) * M / expr1) ** 4
metric = diag(
    -(1 - Rational(1, 2) * M / expr1) ** 2 / (1 + Rational(1, 2) * M / expr1) ** 2,
    expr2,
    expr2,
    expr2,
)



"""
Name: Kerr
References: Allen, Gen. Rel. Grav., v26, p21, (1994)
Coordinates: Cartesian
Symmetry: Axial
"""
from sympy import symbols, zeros

coords = symbols('t x y z', real=True)
variables = symbols('a M', constant=True)
functions = ()
t, x, y, z = coords
a, M = variables
expr = x ** 2 + a ** 2 * y ** 2
metric = zeros(4)
metric[0, 0] = -(a ** 2 * y ** 2 + x ** 2 - 2 * M * x) / expr
metric[1, 1] = expr / (x ** 2 - 2 * M * x + a ** 2)
metric[2, 2] = -expr / (y ** 2 - 1)
metric[3, 3] = (
    -(
        x ** 4
        + x ** 2 * a ** 2
        + 2 * a ** 2 * M * x
        + a ** 2 * y ** 2 * x ** 2
        - 2 * a * y ** 2 * M * x
        + a ** 4 * y ** 2
    )
    * (y ** 2 - 1)
    / expr
)
metric[0, 3] = metric[3, 0] = 2 * a * M * x * (y ** 2 - 1) / expr



from collections import namedtuple
from pathlib import Path

from sympy import Array, Pow, Rational, S, ones, tensorproduct, zeros
from sympy.tensor.tensor import TensorIndexType

from .partial import PartialDerivative, CovariantHead
from .tensor import AbstractTensor, Tensor, expand_array, indices


class Metric(AbstractTensor, TensorIndexType):
    """
    Class representing a tensor that raises and lowers indices.
    """

    # This object allows for having the metric be represented by the
    # same symbol as the tensor it associates with.
    _MetricId = namedtuple('MetricId', ['name', 'antisym'])

    is_Metric = True
    _christoffel = None
    _riemann = None
    _ricci_tensor = None
    _ricci_scalar = None
    _weyl = None
    _einstein = None

    def __new__(cls, symbol, coords, matrix, **kwargs):
        """
        Create a new Metric object.

        Parameters
        ----------
        symbol : str
            Name of the tensor and the symbol to denote it by when printed.
        coords : iterable
            List of ~sympy.Symbol objects to denote the coordinates by which
            derivatives are taken with respect to.
        matrix : (list, tuple, ~sympy.Matrix, ~sympy.Array)
            Matrix representation of the tensor to be used in substitution.
            Can be of any type that is acceptable by ~sympy.Array.

        Examples
        --------
        >>> from sympy import diag, symbols
        >>> from riccipy import Metric, indices, expand_array
        >>> t, x, y, z = symbols('t x y z')
        >>> eta = Metric('eta', [t, x, y, z], diag(1, -1, -1, -1))
        >>> mu, nu = indices('mu nu', eta)
        >>> expr = eta(mu, nu) * eta(-mu, -nu)
        >>> expand_array(expr)
        4
        """
        array = Array(matrix)
        if array.rank() != 2 or array.shape[0] != array.shape[1]:
            raise ValueError(
                'matrix must be square, received matrix of shape {}'.format(array.shape)
            )
        obj = TensorIndexType.__new__(
            cls,
            symbol,
            metric=cls._MetricId(symbol, False),
            dim=array.shape[0],
            dummy_fmt=symbol,
            **kwargs
        )
        obj = AbstractTensor.__new__(cls, obj, array)
        obj.metric = Tensor(obj.name, array, obj, covar=(-1, -1))
        obj.coords = tuple(coords)
        obj._repl[obj] = array
        return obj

    def __getattr__(self, attr):
        if hasattr(self.metric, attr):
            return getattr(self.metric, attr)
        return TensorIndexType.__getattribute__(self, attr)

    def __call__(self, *args):
        return self.metric(*args)

    def __getitem__(self, keys):
        return self.metric.__getitem__(keys)

    def set_variables(self, sub_dict):
        self.metric.set_variables(sub_dict)
        self._array = self._array.subs(sub_dict)
        self._repl[self] = self._array
        self._christoffel = None
        self._riemann = None
        self._ricci_tensor = None
        self._ricci_scalar = None
        self._weyl = None
        self._einstein = None

    def density(self, weight=S.One):
        return Pow(abs(self.determinant), Rational(weight, 2))

    @property
    def determinant(self):
        return self.as_matrix().det()

    @property
    def partial(self):
        return PartialDerivative(self)

    @property
    def covariant(self):
        return CovariantHead(self)

    @property
    def christoffel(self):
        r"""
        Returns the Christoffel symbols using the formula:
        \Gamma^\sigma_{\mu\nu} =
        \frac{1}{2} g^{\sigma\rho} (\partial_\mu g_{\nu\rho} + \partial_\nu g_{\rho\mu} - \partial_\rho g_{\mu\nu})
        """
        if self._christoffel is None:
            mu, nu, si, rh = indices('mu nu sigma rho', self)
            d = self.partial
            g = self.metric
            gamma = (
                Rational(1, 2)
                * g(si, rh)
                * (d(-mu) * g(-nu, -rh) + d(-nu) * g(-rh, -mu) - d(-rh) * g(-mu, -nu))
            )
            syms = expand_array(gamma, [si, -mu, -nu])
            self._christoffel = Tensor('\u0393', syms, self, covar=(1, -1, -1))
        return self._christoffel

    @property
    def riemann(self):
        r"""
        Returns the Riemann curvature tensor using the formula:
        R^\rho_{\sigma\mu\nu} =
        \partial_\mu \Gamma^\rho_{\nu\sigma} - \partial_\nu \Gamma^\rho_{\mu\sigma}
        + \Gamma^\rho_{\mu\lambda} \Gamma^\lambda_{\nu\sigma} - \Gamma^\rho_{\nu\lambda} \Gamma^\lambda_{\mu\sigma}
        """
        if self._riemann is None:
            mu, nu, si, rh, la = indices('mu nu sigma rho lambda', self)
            d = self.partial
            g = self.metric
            gamma = self.christoffel
            R = (
                d(-mu) * gamma(rh, -nu, -si)
                - d(-nu) * gamma(rh, -mu, -si)
                + gamma(rh, -mu, -la) * gamma(la, -nu, -si)
                - gamma(rh, -nu, -la) * gamma(la, -mu, -si)
            )
            res = expand_array(R, [rh, -si, -mu, -nu])
            self._riemann = Tensor(
                'R', res, self, symmetry=(2, 2), covar=(1, -1, -1, -1)
            )
        return self._riemann

    @property
    def ricci_tensor(self):
        r"""
        Returns the Ricci tensor using the formula:
        R_{\mu\nu} = R^\sigma_{\mu\sigma\nu}
        """
        if self._ricci_tensor is None:
            mu, nu, si = indices('mu nu sigma', self)
            R = self.riemann
            res = expand_array(R(si, -mu, -si, -nu), [-mu, -nu])
            self._ricci_tensor = Tensor('R', res, self, covar=(-1, -1))
        return self._ricci_tensor

    @property
    def ricci_scalar(self):
        r"""
        Returns the Ricci scalar using the formula:
        R = R^\mu_\mu
        """
        if self._ricci_scalar is None:
            mu, nu = indices('mu nu', self)
            g = self.metric
            RR = self.ricci_tensor
            res = expand_array(g(-mu, -nu) * RR(mu, nu))
            self._ricci_scalar = res
        return self._ricci_scalar

    @property
    def weyl(self):
        r"""
        Returns the Weyl conformal tensor using the formula:
        C_{\rho\sigma\mu\nu} =
        R_{\rho\sigma\mu\nu} - \frac{2}{(n - 2)} (g_{\rho[\mu} R_{\nu]\sigma} - g_{\sigma[\mu} R_{\nu]\rho})
        + \frac{2}{(n - 1)(n - 2)} g_{\rho[\mu} g_{\nu]\sigma} R
        """
        if self._weyl is None:
            n = self.dim
            if n < 3:
                raise ValueError(
                    'the Weyl tensor is only defined in dimensions of 3 or more. {} is of dimension {}'.format(
                        self, n
                    )
                )
            elif n == 3:
                res = tensorproduct(zeros(3, 3), zeros(3, 3))
                self._weyl = Tensor(
                    'C', res, self, symmetry=(2, 2), covar=(1, -1, -1, -1)
                )
                return self._weyl
            c1 = Rational(1, n - 2)
            c2 = Rational(1, (n - 2) * (n - 1))
            mu, nu, si, rh = indices('mu nu sigma rho', self)
            R = self.riemann
            RR = self.ricci_tensor
            RRR = self.ricci_scalar
            g = self.metric
            C = (
                R(rh, -si, -mu, -nu)
                - c1
                * (
                    g(rh, -mu) * RR(-nu, -si)
                    - g(rh, -nu) * RR(-mu, -si)
                    + g(-si, -nu) * RR(-mu, rh)
                    - g(-si, -mu) * RR(-nu, rh)
                )
                + c2 * (g(rh, -mu) * g(-nu, -si) - g(rh, -nu) * g(-mu, -si)) * RRR
            )
            res = expand_array(C, [rh, -si, -mu, -nu])
            self._weyl = Tensor(
                'C', res, self, symmetry=(2, 2), covar=(1, -1, -1, -1)
            )
        return self._weyl

    @property
    def einstein(self):
        r"""
        Returns the Einstein tensor using the formula:
        G_{\mu\nu} = R_{\mu\nu} - \frac{1}{2} R g_{\mu\nu}
        """
        if self._einstein is None:
            mu, nu = indices('mu nu', self)
            g = self.metric
            R = self.ricci_tensor
            RR = self.ricci_scalar
            res = expand_array(R(-mu, -nu) - Rational(1, 2) * RR * g(-mu, -nu))
            self._einstein = Tensor('G', res, self, covar=(-1, -1))
        return self._einstein


class SpacetimeMetric(Metric):
    """
    Class representing psuedo-Riemannian metrics.
    """

    is_Spacetime = True

    def __new__(cls, symbol, coords, matrix, timelike=True, **kwargs):
        obj = super().__new__(cls, symbol, coords, matrix, **kwargs)
        if obj.dim > 4:
            raise ValueError('metrics on spacetime must be at most 4-dimensional')
        obj.is_timelike = timelike
        obj.is_spacelike = not timelike
        return obj

    def reverse_signature(self):
        self._array *= -1
        self._replacement_dict = {self: self._array}
        self.is_timelike = not self.is_timelike
        self.is_spacelike = not self.is_spacelike
        return self.signature

    @property
    def signature(self):
        sign = -1 if self.is_timelike else 1
        sig = sign * ones(1, self.dim)
        sig[0] *= -1
        return tuple(sig)


def load_metric(symbol, name, coords=None, notes=None, timelike=False):
    from .metrics import data
    spacetime = data(name, coords=coords, notes=notes)
    if isinstance(spacetime, list):
        spacetime = spacetime[0]
    metric = SpacetimeMetric(symbol, spacetime['coords'], spacetime['metric'], timelike=False)
    if timelike:
        metric.reverse_signature()
    variables = spacetime['variables']
    functions = spacetime['functions']
    return (metric, variables, functions)



from collections import defaultdict

from sympy import Array, preorder_traversal, simplify, symbols
from sympy.core.compatibility import string_types
from sympy.core.decorators import call_highest_priority
from sympy.tensor.array import permutedims, tensorcontraction, tensorproduct
from sympy.tensor.tensor import (
    TensMul,
    Tensor as SympyTensor,
    TensorHead,
    TensorIndex,
    TensorManager,
    TensorHead,
    TensorSymmetry,
    get_symmetric_group_sgs,
)


class Repl(dict):
    def getkey(self, tensor):
        item = tensor if (tensor.is_TensorHead or tensor.is_Metric) else tensor.args[0]
        for key in self.keys():
            if (key == item) or (key.args[0] == item):
                return key
        return None

    def haskey(self, tensor):
        key = self.getkey(tensor)
        return True if key is not None else False

    def setitem(self, tensor, array):
        key = self.getkey(tensor)
        if key is not None:
            return super().__setitem__(key, array)
        raise KeyError(str(tensor))

    def __setitem__(self, tensor, array):
        if self.haskey(tensor):
            return self.setitem(tensor, array)
        return super().__setitem__(tensor, array)


class AbstractTensor(object):
    """
    Wrapper class for sympy Array with attributes used for identification.
    """

    is_Tensor = True
    is_TensorHead = False
    is_Metric = False
    is_Spacetime = False
    is_TensorDerivative = False
    _array = None
    _inverse = None

    def __new__(cls, obj, matrix):
        obj._array = matrix
        obj._repl = Repl()
        return obj

    def as_matrix(self):
        """
        Return the data stored in the tensor as an instance of sympy.Matrix.

        Notes
        -----
        The ``tomatrix`` attribute required by this method will throw an error
        if the tensor is not of rank 2.
        """
        return self._array.tomatrix()

    def as_array(self):
        """
        Return the data stored in the tensor as an instance of sympy.Array.
        """
        return self._array.copy()

    def as_inverse(self):
        if self._inverse is None:
            self._inverse = Array(self.as_matrix().inv())
        return self._inverse


class IndexedTensor(AbstractTensor, SympyTensor):
    """
    Class representing a Tensor that has been evaluated with indices.

    Used as the base object for creating algebraic expressions with tensors.
    Inherited are methods such as __mul__ to allow sympy to manage dummy
    indices when multiplied with other tensors.

    Generated when a Tensor is called as a function with indices as arguments.
    """
    _op_priority = 11

    def __new__(cls, tensor, indices, **kwargs):
        obj = SympyTensor.__new__(cls, tensor, indices, **kwargs)
        array = tensor.covariance_transform(*indices)
        return AbstractTensor.__new__(cls, obj, array)

    @call_highest_priority('__rmul__')
    def __mul__(self, other):
        return SympyTensor.__mul__(self, other)

    @call_highest_priority('__mul__')
    def __rmul__(self, other):
        return SympyTensor.__rmul__(self, other)


class Tensor(AbstractTensor, TensorHead):
    """
    Class for representing a Sympy TensorHead object that have an associated
    array of data elements/expressions to be substituted when requested.
    """

    is_TensorHead = True

    def __new__(cls, symbol, matrix, metric, **kwargs):
        """
        Create a new Tensor object.

        Parameters
        ----------
        symbol : str
            Name of the tensor and the symbol to denote it by when printed.
        matrix : (list, tuple, sympy.Matrix, sympy.Array)
            Matrix representation of the tensor to be used in substitution.
            Can be of any type that is acceptable by sympy.Array.
        metric : Metric
            Classify the tensor as being defined in terms of a metric.

        Notes
        -----
        If the parameter ``symmetry`` is passed, the tensor object will defined
        using a specific symmetry. Example values are (see sympy documentation
        for the function ``tensorsymmetry``):
        ``[[1]]``         vector
        ``[[1]*n]``       symmetric tensor of rank ``n``
        ``[[n]]``         antisymmetric tensor of rank ``n``
        ``[[2, 2]]``      monoterm slot symmetry of the Riemann tensor
        ``[[1],[1]]``     vector*vector
        ``[[2],[1],[1]]`` (antisymmetric tensor)*vector*vector

        Additionally, the parameter ``covar`` indicates that the passed array
        corresponds to the covariance of the tensor it is intended to describe.

        Lastly, the parameter ``comm`` is used to indicate what commutation
        group the tensor belongs to. In other words, it describes what other
        types of tensors the one being created is allowed to commute with.
        There are three commutation groups: ``general`` for ordinary tensors,
        ``metric`` for metric tensors, and ``partial`` for partial derivatives.

        Examples
        --------
        >>> from sympy import diag, symbols
        >>> from riccipy import Metric, Tensor, indices, expand_array
        >>> E1, E2, E3, B1, B2, B3 = symbols('E1:4 B1:4')
        >>> em = [[0, -E1, -E2, -E3],
                  [E1, 0, -B3, B2],
                  [E2, B3, 0, -B1],
                  [E3, -B2, B1, 0]]
        >>> t, x, y, z = symbols('t x y z')
        >>> eta = Metric('eta', [t, x, y, z], diag(1, -1, -1, -1))
        >>> F = Tensor('F', em, eta, symmetry=[[2]])
        >>> mu, nu = indices('mu nu', eta)
        >>> expr = F(mu, nu) + F(nu, mu)
        >>> expand_array(expr)
        0
        >>> expr = F(mu, nu) * F(-mu, -nu)
        >>> expand_array(expr)
        2*B_1**2 + 2*B_2**2 + 2*B_3**2 - 2*E_1**2 - 2*E_2**2 - 2*E_3**2
        """
        array = Array(matrix)
        sym = kwargs.pop('symmetry', array.rank() * [1])
        sym = TensorSymmetry.direct_product(*sym)
        comm = kwargs.pop('comm', 'general')
        covar = tuple(kwargs.pop('covar', array.rank() * [1]))
        if len(covar) != array.rank():
            raise ValueError(
                'covariance signature {} does not match tensor rank {}'.format(
                    covar, array.rank()
                )
            )

        obj = TensorHead.__new__(cls, symbol, array.rank() * [metric], sym, comm=comm, **kwargs)
        obj = AbstractTensor.__new__(cls, obj, array)
        # resolves a bug with pretty printing.
        obj.__class__.__name__ = 'TensorHead'
        obj.covar = covar
        idxs = obj._dummy_idxs()
        obj._repl[obj(*idxs)] = array
        return obj

    def __repr__(self):
        return self._print()

    def __str__(self):
        return self.__repr__()

    def __call__(self, *args, **kwargs):
        new = IndexedTensor(self, args, **kwargs)
        return new.doit()

    def __getitem__(self, keys):
        return self._array.__getitem__(keys)

    def _dummy_idxs(self):
        count = defaultdict(int)

        def dummy_fmt_gen(idxtype):
            # generate a generic index for the entry in replacement dictionary.
            fmt = idxtype.dummy_fmt
            n = count[idxtype]
            count[idxtype] += 1
            return fmt % n

        idx_names = map(dummy_fmt_gen, self.index_types)
        idx_generator = map(Index, idx_names, self.index_types)
        idxs = [
            idx if self.covar[pos] > 0 else -idx for pos, idx in enumerate(idx_generator)
        ]
        return idxs

    def set_variables(self, sub_dict):
        self._array = self._array.subs(sub_dict)
        idxs = self._dummy_idxs()
        self._repl[self(*idxs)] = self._array

    def covariance_transform(self, *indices):
        """
        Return the array associated with this tensor with indices set according
        to arguments.

        Parameters
        ----------
        indices : TensorIndex
            Defines the covariance and contravariance of the returned array.

        Examples
        --------
        >>> from sympy import diag, symbols, sin
        >>> from riccipy import Metric, indices
        >>> t, r, th, ph = symbols('t r theta phi')
        >>> schwarzschild = diag(1-1/r, -1/(1-1/r), -r**2, -r**2*sin(th)**2)
        >>> g = Metric('g', [t, r, th, ph], schwarzschild)
        >>> mu, nu = indices('mu nu', g)
        >>> g.covariance_transform(mu, nu)
        [[1/(1 - 1/r), 0, 0, 0], [0, 1 - 1/r, 0, 0], [0, 0, -1/r**2, 0], [0, 0, 0, -1/(r**2*sin(theta)**2)]]
        """
        array = self.as_array()
        for pos, idx in enumerate(indices):
            if idx.is_up ^ (self.covar[pos] > 0):
                if idx.is_up:
                    metric = idx.tensor_index_type.metric.as_inverse()
                else:
                    metric = idx.tensor_index_type.metric.as_array()
                new = tensorcontraction(tensorproduct(metric, array), (1, 2 + pos))
                permu = list(range(len(indices)))
                permu[0], permu[pos] = permu[pos], permu[0]
                array = permutedims(new, permu)
        return array

    def simplify(self):
        """
        Replace the stored array associated with this tensor with a simplified
        version. This method also replaces the entry in the replacement dictionary.
        """
        array = self.as_array().applyfunc(simplify)
        self._array = array
        self._repl.setitem(self, array)
        return array


class Index(TensorIndex):
    """
    Class for a symbolic representation of a tensor index with respect to a metric.
    """

    def __new__(cls, symbol, metric, is_up=True, **kwargs):
        return super().__new__(cls, symbol, metric, is_up=is_up, **kwargs)

    def __neg__(self):
        return Index(self.name, self.tensor_index_type, (not self.is_up))


def expand_array(expr, idxs=None):
    """
    Evaluate a tensor expression and return the result as an array.

    Parameters
    ----------
    expr : TensExpr
        Symbolic expression of tensors.
    idxs : TensorIndex
        Indices that encode the covariance and contravariance of the result.
    """
    repl = Repl()

    for arg in preorder_traversal(expr):
        if isinstance(arg, AbstractTensor):
            repl.update(arg._repl)
            for metric in arg.index_types:
                repl.update(metric._repl)

    if idxs is None:
        idxs = TensMul(expr).get_free_indices()
    return expr.replace_with_arrays(repl, idxs)


def expand_tensor(symbol, expr, metric, idxs=None, **kwargs):
    """
    Evaluate a tensor expression and return the result as a tensor.

    Parameters
    ----------
    symbol : str
        Name of the tensor and the symbol to denote it by when printed.
    expr : TensExpr
        Symbolic expression of tensors.
    metric : Metric
        Classify the tensor as being defined in terms of a metric.
    idxs : TensorIndex
        Indices that encode the covariance and contravariance of the result.
    """
    result = expand_array(expr, idxs)
    if not isinstance(result, Array):
        return result
    if idxs is None:
        idxs = TensMul(expr).get_free_indices()
    covar = [1 if idx.is_up else -1 for idx in idxs]
    return Tensor(symbol, result, metric, covar=covar, **kwargs)


def indices(s, metric, is_up=True):
    """
    Create indices using a method similar to sympy.symbols.
    """
    if isinstance(s, string_types):
        a = [x.name for x in symbols(s, seq=True)]
    else:
        raise ValueError(
            'expected a string, received object of type {}'.format(type(s))
        )
    idxs = [Index(idx, metric, is_up) for idx in a]
    if len(idxs) == 1:
        return idxs[0]
    return idxs


# metric tensors and general tensors commute with each other and themselves.
TensorManager.set_comm('general', 'general', 0)
# partial derivatives only commute with themselves.
TensorManager.set_comm('partial', 'partial', 0)



"""
Name: Schwarzschild
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
Notes: Isotropic Coordinates
"""
from sympy import Rational, diag, sin, symbols

coords = symbols('t r theta phi', real=True)
variables = symbols('M', constant=True)
functions = ()
t, r, th, ph = coords
M = variables
expr = (1 + Rational(1, 2) * M / r) ** 4
metric = diag(
    -(1 - Rational(1, 2) * M / r) ** 2 / (1 + Rational(1, 2) * M / r) ** 2,
    expr,
    expr * r ** 2,
    expr * r ** 2 * sin(th) ** 2,
)



"""
Name: Levi-Civita Vacuum
References: Stephani (20.8) p221
Coordinates: Cylindrical
Notes: m = 2
"""
from sympy import diag, symbols

coords = symbols('t rho phi z', real=True)
variables = ()
functions = ()
t, rh, ph, z = coords
metric = diag(-rh ** 4, rh ** 4, 1 / rh ** 2, rh ** 4)



"""
Name: Tolman
References: Tolman, Phys. Rev., v55, p363-373, (1939)
Coordinates: Spherical
Notes: Type VI
"""
from sympy import diag, sin, symbols

coords = symbols('t r theta phi', real=True)
variables = symbols('A B n', constant=True)
functions = ()
t, r, th, ph = coords
A, B, n = variables
metric = diag(
    -(A * r ** (1 - n) - B * r ** (1 + n)) ** 2,
    2 - n ** 2,
    r ** 2,
    r ** 2 * sin(th) ** 2,
)



from setuptools import setup

long_description = """RicciPy is a Python tensor algebra calculator for
symbolically manipulating exact solutions to the Einstein Field Equations."""

version = '0.2a'
version_ref = '1'

setup(
    name='riccipy',
    author='Calvin Jay Ross',
    author_email='calvinjayross@gmail.com',
    description='A tensor algebra calculator for General Relativity',
    long_description=long_description,
    url='https://github.com/cjayross/riccipy',
    download_url='https://github.com/cjayross/riccipy/archive/v' + version + '.tar.gz',
    license='MIT',
    version=version + version_ref,
    packages=['riccipy', 'riccipy.metrics', 'riccipy.tests'],
    install_requires=open('requirements.txt').read().split('\n')[:-1],
    keywords='general relativity physics math cas',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



"""
Name: Beckers, Sinzinkayo, and Demaret
References: Beckers et al., Phys. Rev. D, v30, p1846, (1984)
Coordinates: Cartesian
"""
from sympy import diag, symbols

coords = symbols('t x y z', real=True)
variables = symbols('k d m', constant=True)
functions = ()
t, x, y, z = coords
k, d, m = variables
expr = (k * x + d) ** (2 * m)
metric = diag(-expr, expr, expr, expr)



"""
Name: Levi-Civita
References: Stephani (Table 16.2) p188
Coordinates: Cylindrical
Notes: Class A2
"""
from sympy import diag, sinh, symbols

coords = symbols('t r phi z', real=True)
variables = symbols('M', constant=True)
functions = ()
t, r, ph, z = coords
M = variables
metric = diag(-(2 * M / z - 1), z ** 2, z ** 2 * sinh(r) ** 2, 1 / (2 * M / z - 1))



"""
Name: Bianchi
References: Stephani (13.49) p162
Coordinates: Cartesian
Symmetry: Planar
Notes: Bianchi I
"""
from sympy import Function, diag, exp, symbols

coords = symbols('t x y z', real=True)
variables = ()
functions = symbols('alpha beta', cls=Function)
t, x, y, z = coords
al, be = functions
metric = diag(-1, exp(2 * be(t)), exp(2 * be(t)), exp(2 * al(t)))



"""
RicciPy
=======

Python library for tensor manipulation in General Relativity
"""

__version__ = "0.1-alpha"
from .metric import Metric, SpacetimeMetric, load_metric
from .numerical import lambdify_tensor
from .tensor import Index, Tensor, expand_array, expand_tensor, indices



"""
Name: Kerr Newman
References:
    - Newman, J. Math. Phys., v6, p918, (1965)
    - Stephani (19.19) p213
Coordinates: Boyer Lindquist
Symmetry: Axial
"""
from sympy import cos, sin, symbols, zeros

coords = symbols('t r theta phi', real=True)
variables = symbols('a M Q', constant=True)
functions = ()
t, r, th, ph = coords
a, M, Q = variables
expr = a ** 2 * cos(th) ** 2 + r ** 2
metric = zeros(4)
metric[0, 0] = -(a ** 2 - sin(th) ** 2 * a ** 2 + r ** 2 - (2 * M * r - Q ** 2)) / expr
metric[1, 1] = expr / (r ** 2 - 2 * M * r + Q ** 2 + a ** 2)
metric[2, 2] = expr
metric[3, 3] = (
    (
        -a ** 4 * sin(th) ** 2
        + a ** 4
        + 2 * r ** 2 * a ** 2
        - a ** 2 * sin(th) ** 2 * r ** 2
        + 2 * a ** 2 * sin(th) ** 2 * M * r
        - a ** 2 * sin(th) ** 2 * Q ** 2
        + r ** 4
    )
    * sin(th) ** 2
    / expr
)
metric[0, 3] = metric[3, 0] = (
    -(2 * a * sin(th) ** 2 * M * r - a * sin(th) ** 2 * Q ** 2) / expr
)



"""
Name: C-metric
References: Stephani (Table 16.2) p188
"""
from sympy import Function, diag, symbols

coords = symbols('t x y phi', real=True)
variables = ()
functions = symbols('f h', cls=Function)
t, x, y, ph = coords
f, h = functions
metric = diag(
    -h(y) / (x + y) ** 2,
    1 / ((x + y) ** 2 * f(x)),
    1 / ((x + y) ** 2 * h(y)),
    f(x) / (x + y) ** 2,
)



"""
Name: Levi-Civita
References: Stephani (Table 16.2) p188
Coordinates: Cylindrical
Notes: Class B3
"""
from sympy import diag, symbols

coords = symbols('t r phi z', real=True)
variables = ()
functions = ()
t, r, ph, z = coords
metric = diag(-r ** 2 * z ** 2, z ** 2, 1 / z, z)



"""
Name: Einstein
References: Stephani (10.23a) p122
Coordinates: Cartesian
Symmetry: Static
"""
from sympy import diag, sqrt, symbols

coords = symbols('t x y z', real=True)
variables = symbols('K', constant=True)
functions = ()
t, x, y, z = coords
K = variables
expr = 1 / (1 + sqrt(x ** 2 + y ** 2 + z ** 2) / (4 * K ** 2)) ** 2
metric = diag(-1, expr, expr, expr)



"""
Name: Anti-de Sitter
References: Hawking and Ellis (5.9) p131
Symmetry: Maximal
"""
from sympy import cos, diag, sin, sinh, symbols

coords = symbols('t chi theta phi', real=True)
variables = ()
functions = ()
t, ch, th, ph = coords
metric = diag(
    -1,
    cos(t) ** 2,
    cos(t) ** 2 * sinh(ch) ** 2,
    cos(t) ** 2 * sinh(ch) ** 2 * sin(th) ** 2,
)
